@@ -9,7 +9,6 @@
 -- Stability   :  experimental
 -- Portability :  portable
 -----------------------------------------------------------------------------
-
 module Haddock.Interface.Create (createInterface) where
 
 
@@ -30,61 +29,73 @@ import Control.Monad
 import qualified Data.Traversable as Traversable
 
 import GHC hiding (flags)
+import HscTypes
 import Name
 import Bag
 import RdrName (GlobalRdrEnv)
 
 
--- | Process the data in the GhcModule to produce an interface.
+-- | Process the data in a GhcModule to produce an interface.
 -- To do this, we need access to already processed modules in the topological
--- sort. That's what's in the module map.
-createInterface :: GhcModule -> [Flag] -> ModuleMap -> InstIfaceMap
+-- sort. That's what's in the interface map.
+createInterface :: TypecheckedModule -> [Flag] -> IfaceMap -> InstIfaceMap
                 -> ErrMsgGhc Interface
-createInterface ghcMod flags modMap instIfaceMap = do
+createInterface tm flags modMap instIfaceMap = do
 
-  let mdl = ghcModule ghcMod
+  let ms                  = pm_mod_summary . tm_parsed_module $ tm
+      mi                  = moduleInfo tm
+      mdl                 = ms_mod ms
+      dflags              = ms_hspp_opts ms
+      instances           = modInfoInstances mi
+      exportedNames       = modInfoExports mi
+      -- XXX: confirm always a Just.
+      Just (group_, _, optExports, optDocHeader) = renamedSource tm
 
   -- The pattern-match should not fail, because createInterface is only
   -- done on loaded modules.
   Just gre <- liftGhcToErrMsgGhc $ lookupLoadedHomeModuleGRE (moduleName mdl)
 
-  opts0 <- liftErrMsg $ mkDocOpts (ghcMbDocOpts ghcMod) flags mdl
+  opts0 <- liftErrMsg $ mkDocOpts (haddockOptions dflags) flags mdl
   let opts
         | Flag_IgnoreAllExports `elem` flags = OptIgnoreExports : opts0
         | otherwise = opts0
 
-  (info, mbDoc)    <- liftErrMsg $ lexParseRnHaddockModHeader
-                                       gre (ghcMbDocHdr ghcMod)
-  decls0           <- liftErrMsg $ declInfos gre (topDecls (ghcGroup ghcMod))
+  (info, mbDoc)    <- liftErrMsg $ lexParseRnHaddockModHeader dflags gre optDocHeader
+  decls0           <- liftErrMsg $ declInfos dflags gre (topDecls group_)
 
-  let instances      = ghcInstances ghcMod
-      localInsts     = filter (nameIsLocalOrFrom mdl . getName) instances
+  let localInsts     = filter (nameIsLocalOrFrom mdl . getName) instances
       declDocs       = [ (decl, doc) | (L _ decl, (Just doc, _), _) <- decls0 ]
       instanceDocMap = mkInstanceDocMap localInsts declDocs
 
       decls         = filterOutInstances decls0
       declMap       = mkDeclMap decls
-      exports       = fmap (reverse . map unLoc) (ghcMbExports ghcMod)
+      exports       = fmap (reverse . map unLoc) optExports
       ignoreExps    = Flag_IgnoreAllExports `elem` flags
-      exportedNames = ghcExportedNames ghcMod
 
   liftErrMsg $ warnAboutFilteredDecls mdl decls0
 
-  exportItems <- mkExportItems modMap mdl gre (ghcExportedNames ghcMod) decls declMap
-                               opts exports ignoreExps instances instIfaceMap
+  exportItems <- mkExportItems modMap mdl gre exportedNames decls declMap
+                               opts exports ignoreExps instances instIfaceMap dflags
 
   let visibleNames = mkVisibleNames exportItems opts
 
+  -- measure haddock documentation coverage.
+  let
+    prunedExportItems0 = pruneExportItems exportItems
+    haddockable = 1 + length exportItems -- module + exports
+    haddocked = (if isJust mbDoc then 1 else 0) + length prunedExportItems0
+    coverage = (haddockable, haddocked)
+
   -- prune the export list to just those declarations that have
   -- documentation, if the 'prune' option is on.
   let
     prunedExportItems
-      | OptPrune `elem` opts = pruneExportItems exportItems
+      | OptPrune `elem` opts = prunedExportItems0
       | otherwise = exportItems
 
   return Interface {
     ifaceMod             = mdl,
-    ifaceOrigFilename    = ghcFilename ghcMod,
+    ifaceOrigFilename    = msHsFilePath ms,
     ifaceInfo            = info,
     ifaceDoc             = mbDoc,
     ifaceRnDoc           = Nothing,
@@ -97,7 +108,8 @@ createInterface ghcMod flags modMap instIfaceMap = do
     ifaceDeclMap         = declMap,
     ifaceSubMap          = mkSubMap declMap exportedNames,
     ifaceInstances       = instances,
-    ifaceInstanceDocMap  = instanceDocMap
+    ifaceInstanceDocMap  = instanceDocMap,
+    ifaceHaddockCoverage = coverage
   }
 
 
@@ -168,22 +180,22 @@ mkDeclMap decls = Map.fromList . concat $
   , not (isDocD d), not (isInstD d) ]
 
 
-declInfos :: GlobalRdrEnv -> [(Decl, MaybeDocStrings)] -> ErrMsgM [DeclInfo]
-declInfos gre decls =
+declInfos :: DynFlags -> GlobalRdrEnv -> [(Decl, MaybeDocStrings)] -> ErrMsgM [DeclInfo]
+declInfos dflags gre decls =
   forM decls $ \(parent@(L _ d), mbDocString) -> do
-            mbDoc <- lexParseRnHaddockCommentList NormalHaddockComment
+            mbDoc <- lexParseRnHaddockCommentList dflags NormalHaddockComment
                        gre mbDocString
             fnArgsDoc <- fmap (Map.mapMaybe id) $
                 Traversable.forM (getDeclFnArgDocs d) $
-                \doc -> lexParseRnHaddockComment NormalHaddockComment gre doc
+                \doc -> lexParseRnHaddockComment dflags NormalHaddockComment gre doc
 
             let subs_ = subordinates d
             subs <- forM subs_ $ \(subName, mbSubDocStr, subFnArgsDocStr) -> do
-                mbSubDoc <- lexParseRnHaddockCommentList NormalHaddockComment
+                mbSubDoc <- lexParseRnHaddockCommentList dflags NormalHaddockComment
                               gre mbSubDocStr
                 subFnArgsDoc <- fmap (Map.mapMaybe id) $
                   Traversable.forM subFnArgsDocStr $
-                  \doc -> lexParseRnHaddockComment NormalHaddockComment gre doc
+                  \doc -> lexParseRnHaddockComment dflags NormalHaddockComment gre doc
                 return (subName, (mbSubDoc, subFnArgsDoc))
 
             return (parent, (mbDoc, fnArgsDoc), subs)
@@ -267,13 +279,17 @@ filterOutInstances = filter (\(L _ d, _, _) -> not (isInstD d))
 -- bindings from an 'HsGroup'.
 declsFromGroup :: HsGroup Name -> [Decl]
 declsFromGroup group_ =
-  mkDecls hs_tyclds  TyClD    group_ ++
-  mkDecls hs_derivds DerivD   group_ ++
-  mkDecls hs_defds   DefD     group_ ++
-  mkDecls hs_fords   ForD     group_ ++
-  mkDecls hs_docs    DocD     group_ ++
-  mkDecls hs_instds  InstD    group_ ++
-  mkDecls (typesigs . hs_valds) SigD group_
+#if MIN_VERSION_ghc(7,0,2)
+  mkDecls (concat . hs_tyclds)  TyClD  group_ ++
+#else
+  mkDecls hs_tyclds             TyClD  group_ ++
+#endif
+  mkDecls hs_derivds            DerivD group_ ++
+  mkDecls hs_defds              DefD   group_ ++
+  mkDecls hs_fords              ForD   group_ ++
+  mkDecls hs_docs               DocD   group_ ++
+  mkDecls hs_instds             InstD  group_ ++
+  mkDecls (typesigs . hs_valds) SigD   group_
   where
     typesigs (ValBindsOut _ sigs) = filter isVanillaLSig sigs
     typesigs _ = error "expected ValBindsOut"
@@ -420,7 +436,7 @@ finishedDoc d doc rest = (d, docStringToList doc) : rest
 -- We create the export items even if the module is hidden, since they
 -- might be useful when creating the export items for other modules.
 mkExportItems
-  :: ModuleMap
+  :: IfaceMap
   -> Module             -- this module
   -> GlobalRdrEnv
   -> [Name]             -- exported names (orig)
@@ -431,10 +447,11 @@ mkExportItems
   -> Bool               -- --ignore-all-exports flag
   -> [Instance]
   -> InstIfaceMap
+  -> DynFlags
   -> ErrMsgGhc [ExportItem Name]
 
 mkExportItems modMap this_mod gre exported_names decls declMap
-              opts maybe_exps ignore_all_exports _ instIfaceMap
+              opts maybe_exps ignore_all_exports _ instIfaceMap dflags
   | isNothing maybe_exps || ignore_all_exports || OptIgnoreExports `elem` opts
     = everything_local_exported
   | otherwise = liftM concat $ mapM lookupExport (fromJust maybe_exps)
@@ -442,7 +459,7 @@ mkExportItems modMap this_mod gre exported_names decls declMap
 
 
     everything_local_exported =  -- everything exported
-      liftErrMsg $ fullContentsOfThisModule gre decls
+      liftErrMsg $ fullContentsOfThisModule dflags gre decls
 
 
     lookupExport (IEVar x) = declWith x
@@ -451,15 +468,15 @@ mkExportItems modMap this_mod gre exported_names decls declMap
     lookupExport (IEThingWith t _)     = declWith t
     lookupExport (IEModuleContents m)  = fullContentsOf m
     lookupExport (IEGroup lev docStr)  = liftErrMsg $
-      ifDoc (lexParseRnHaddockComment DocSectionComment gre docStr)
+      ifDoc (lexParseRnHaddockComment dflags DocSectionComment gre docStr)
             (\doc -> return [ ExportGroup lev "" doc ])
     lookupExport (IEDoc docStr)        = liftErrMsg $
-      ifDoc (lexParseRnHaddockComment NormalHaddockComment gre docStr)
+      ifDoc (lexParseRnHaddockComment dflags NormalHaddockComment gre docStr)
             (\doc -> return [ ExportDoc doc ])
     lookupExport (IEDocNamed str) = liftErrMsg $
       ifDoc (findNamedDoc str [ unL d | (d,_,_) <- decls ])
             (\docStr ->
-            ifDoc (lexParseRnHaddockComment NormalHaddockComment gre docStr)
+            ifDoc (lexParseRnHaddockComment dflags NormalHaddockComment gre docStr)
                   (\doc -> return [ ExportDoc doc ]))
 
 
@@ -558,51 +575,54 @@ mkExportItems modMap this_mod gre exported_names decls declMap
               -- But I might be missing something obvious.  What's important
               -- /here/ is that we behave reasonably when we run into one of
               -- those exported type-inferenced values.
-              isLocalAndTypeInferenced <- liftGhcToErrMsgGhc $
-                    isLoaded (moduleName (nameModule t))
-              if isLocalAndTypeInferenced
-               then do
-                   -- I don't think there can be any subs in this case,
-                   -- currently?  But better not to rely on it.
-                   let subs = subordinatesWithNoDocs (unLoc hsdecl)
-                   return [ mkExportDecl t (hsdecl, noDocForDecl, subs) ]
-               else
-              -- We try to get the subs and docs
-              -- from the installed interface of that package.
-               case Map.lookup (nameModule t) instIfaceMap of
-                -- It's Nothing in the cases where I thought
-                -- Haddock has already warned the user: "Warning: The
-                -- documentation for the following packages are not
-                -- installed. No links will be generated to these packages:
-                -- ..."
-                -- But I guess it was Cabal creating that warning. Anyway,
-                -- this is more serious than links: it's exported decls where
-                -- we don't have the docs that they deserve!
+              isLocalAndTypeInferenced <- liftGhcToErrMsgGhc $ do
+                    let mdl = nameModule t
+                    if modulePackageId mdl == thisPackage dflags then
+                      isLoaded (moduleName mdl)
+                    else return False
 
-                -- We could use 'subordinates' to find the Names of the subs
-                -- (with no docs). Is that necessary? Yes it is, otherwise
-                -- e.g. classes will be shown without their exported subs.
-                Nothing -> do
-                   liftErrMsg $ tell
-                      ["Warning: Couldn't find .haddock for exported "
-                      ++ exportInfoString]
-                   let subs = subordinatesWithNoDocs (unLoc hsdecl)
-                   return [ mkExportDecl t (hsdecl, noDocForDecl, subs) ]
-                Just iface -> do
-                   let subs = case Map.lookup t (instSubMap iface) of
-                           Nothing -> []
-                           Just x -> x
-                   return [ mkExportDecl t
-                     ( hsdecl
-                     , fromMaybe noDocForDecl $
-                          Map.lookup t (instDocMap iface)
-                     , map (\subt ->
-                              ( subt ,
-                                fromMaybe noDocForDecl $
-                                   Map.lookup subt (instDocMap iface)
-                              )
-                           ) subs
-                     )]
+              if isLocalAndTypeInferenced then do
+                -- I don't think there can be any subs in this case,
+                -- currently?  But better not to rely on it.
+                let subs = subordinatesWithNoDocs (unLoc hsdecl)
+                return [ mkExportDecl t (hsdecl, noDocForDecl, subs) ]
+              else
+                -- We try to get the subs and docs
+                -- from the installed interface of that package.
+                case Map.lookup (nameModule t) instIfaceMap of
+                  -- It's Nothing in the cases where I thought
+                  -- Haddock has already warned the user: "Warning: The
+                  -- documentation for the following packages are not
+                  -- installed. No links will be generated to these packages:
+                  -- ..."
+                  -- But I guess it was Cabal creating that warning. Anyway,
+                  -- this is more serious than links: it's exported decls where
+                  -- we don't have the docs that they deserve!
+
+                  -- We could use 'subordinates' to find the Names of the subs
+                  -- (with no docs). Is that necessary? Yes it is, otherwise
+                  -- e.g. classes will be shown without their exported subs.
+                  Nothing -> do
+                     liftErrMsg $ tell
+                        ["Warning: Couldn't find .haddock for exported "
+                        ++ exportInfoString]
+                     let subs = subordinatesWithNoDocs (unLoc hsdecl)
+                     return [ mkExportDecl t (hsdecl, noDocForDecl, subs) ]
+                  Just iface -> do
+                     let subs = case Map.lookup t (instSubMap iface) of
+                             Nothing -> []
+                             Just x -> x
+                     return [ mkExportDecl t
+                       ( hsdecl
+                       , fromMaybe noDocForDecl $
+                            Map.lookup t (instDocMap iface)
+                       , map (\subt ->
+                                ( subt ,
+                                  fromMaybe noDocForDecl $
+                                     Map.lookup subt (instDocMap iface)
+                                )
+                             ) subs
+                       )]
 
 
     mkExportDecl :: Name -> DeclInfo -> ExportItem Name
@@ -618,7 +638,7 @@ mkExportItems modMap this_mod gre exported_names decls declMap
 
 
     fullContentsOf modname
-      | m == this_mod = liftErrMsg $ fullContentsOfThisModule gre decls
+      | m == this_mod = liftErrMsg $ fullContentsOfThisModule dflags gre decls
       | otherwise =
           case Map.lookup m modMap of
             Just iface
@@ -666,14 +686,14 @@ mkExportItems modMap this_mod gre exported_names decls declMap
 -- (For more information, see Trac #69)
 
 
-fullContentsOfThisModule :: GlobalRdrEnv -> [DeclInfo] -> ErrMsgM [ExportItem Name]
-fullContentsOfThisModule gre decls = liftM catMaybes $ mapM mkExportItem decls
+fullContentsOfThisModule :: DynFlags -> GlobalRdrEnv -> [DeclInfo] -> ErrMsgM [ExportItem Name]
+fullContentsOfThisModule dflags gre decls = liftM catMaybes $ mapM mkExportItem decls
   where
     mkExportItem (L _ (DocD (DocGroup lev docStr)), _, _) = do
-        mbDoc <- lexParseRnHaddockComment DocSectionComment gre docStr
+        mbDoc <- lexParseRnHaddockComment dflags DocSectionComment gre docStr
         return $ fmap (ExportGroup lev "") mbDoc
     mkExportItem (L _ (DocD (DocCommentNamed _ docStr)), _, _) = do
-        mbDoc <- lexParseRnHaddockComment NormalHaddockComment gre docStr
+        mbDoc <- lexParseRnHaddockComment dflags NormalHaddockComment gre docStr
         return $ fmap ExportDoc mbDoc
     mkExportItem (decl, doc, subs) = return $ Just $ ExportDecl decl doc subs []
 
